package mods.immibis.macroblocks;

import java.awt.image.BufferedImage;

import javax.imageio.ImageIO;

public class TextureProcessor {
	static boolean GAMMA_CORRECT = true;
	
	public static BufferedImage load(String name) {
		try {
			return ImageIO.read(TextureProcessor.class.getResource(name+".png"));
		} catch(Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	public static float[] getAvgCol(BufferedImage i) {
		float[] sum = {0, 0, 0};
		for(int x = 0; x < i.getWidth(); x++)
			for(int y = 0; y < i.getHeight(); y++)
				sum = add(sum, getcol(i, x, y));
		int count = i.getWidth()*i.getHeight();
		return mul(sum, 1.0f/count);
	}
	
	private static float[] add(float[] a, float[] b) {
		return new float[] {a[0]+b[0], a[1]+b[1], a[2]+b[2]};
	}
	
	private static float[] avg(float[] a, float[] b) {
		return mul(add(a, b), 0.5f);
	}
	
	private static float[] mul(float[] a, float b) {
		return new float[] {a[0]*b, a[1]*b, a[2]*b};
	}
	
	private static float[] getcol(BufferedImage i, int x, int y) {
		int c = i.getRGB(x, y);
		float[] f = new float[] {((c>>16)&255)/255.0f, ((c>>8)&255)/255.0f, (c&255)/255.0f};
		if(GAMMA_CORRECT)
			for(int k = 0; k < 3; k++)
				f[k] = (float)Math.pow(f[k], 2.2);
		return f;
	}
	
	private static void setcol(BufferedImage i, int x, int y, float[] col) {
		float max = Math.max(Math.max(col[0], col[1]), col[2]);
		if(max > 1) {
			col[0] /= max;
			col[1] /= max;
			col[2] /= max;
		}
		int r = channelToRGB(col[0]);
		int g = channelToRGB(col[1]);
		int b = channelToRGB(col[2]);
		
		i.setRGB(x, y, (r<<16) | (g<<8) | b | 0xFF000000);
	}

	private static int channelToRGB(float f) {
		return Math.min(255, Math.max(0, (int)(Math.pow(f, GAMMA_CORRECT ? 1/2.2 : 1) * 255+0.5f)));
	}
	
	private static BufferedImage reduceVariation(BufferedImage i) {
		return reduceVariation(i, 0.55f);
	}
	
	private static BufferedImage reduceVariation(BufferedImage i, float avgfac) {
		float[] avgcol = getAvgCol(i);
		int w = i.getWidth(), h = i.getHeight();
		for(int x = 0; x < w; x++)
			for(int y = 0; y < h; y++)
				setcol(i,x,y, add(mul(getcol(i, x, y), 1-avgfac), mul(avgcol, avgfac)));
		return i;
	}

	public static BufferedImage doBricksEffect(BufferedImage i, float avgfac, float edgefac, int blurrad, Effect effect) {
		
		if(effect == Effect.PILLAR)
			blur(i, blurrad, 2);
		else if(effect == Effect.PILLAR_HORIZONTAL)
			blur(i, 2, blurrad);
		else
			blur(i, blurrad, blurrad);
		
		i = reduceVariation(i, avgfac);
		
		//float naturalVariance = calcNaturalVariance(i, avg);
		
		//float avgfac = !GAMMA_CORRECT ? 0.7f : 0.7f;
		//float edgefac = !GAMMA_CORRECT ? 1.5f : 2.5f;
		
		// average (col+(avg-col)*avgfac)/avg should be within tolerances
		// X = col/avg+(1-col/avg)*avgfac
		// X = naturalVariance + (1-naturalVariance)*avgfac
		// avgfac = (X - naturalVariance) / (1 - naturalVariance)
		//float avgfac = (1.2f - naturalVariance) / (1 - naturalVariance);
		//System.out.println(avgfac+" "+naturalVariance);
		
		//float avgfac = 1.0f / naturalVariance;
		
		int w = i.getWidth(), h = i.getHeight();
		for(int x = 0; x < w; x++) {
			for(int y = 0; y < h; y++) {
				
				float scalefac = getBricksScaleFactor(edgefac, w, h, x, y, effect);
				
				setcol(i,x,y, mul(getcol(i, x, y), scalefac)); 
			}
		}
		
		return i;
	}

	private static float getBricksScaleFactor(float edgefac, int w, int h, int x, int y, Effect ef) {
		float nedgefac = 1.0f/edgefac;
		
		if(ef == Effect.ORNATE_LARGE_TILES) {
			int ix = (x*16/w)%8, iy = (y*16/h)%8;
			if(ix == 7 || iy == 7)
				return nedgefac;
			if(ix == 0 || iy == 0)
				return edgefac;
			if(ix == 6 || iy == 6)
				return edgefac;
			if(ix == 1 || iy == 1)
				return nedgefac;
			return 1;
		}
		if(ef == Effect.LARGE_TILES || ef == Effect.SMALL_TILES) {
			int size = (ef == Effect.LARGE_TILES ? 8 : 4);
			int ix = (x*16/w)%size, iy = (y*16/h)%size;
			if(ix == 0 || iy == 0)
				return edgefac;
			if(ix == size-1 || iy == size-1)
				return nedgefac;
			return 1;
		}
		if(ef == Effect.PILLAR || ef == Effect.PILLAR_HORIZONTAL) {
			int ix;
			if(ef == Effect.PILLAR)
				ix = (x * 16 / w);
			else
				ix = (y * 16 / h);
			int pos = ix % 4;
			switch(pos) {
			case 0: return 0.84f;
			case 1: return 1.28f;
			case 2: return 1.12f;
			case 3: return 0.72f;
			}
			if((ix + 1) / 2 % 2 == 0)
				return 1;
			return nedgefac;
			
		} else if(ef == Effect.PILLAR_END) {
			int ix = x * 16 / w;
			int iy = y * 16 / h;
			if(ix == 0 || ix == 15 || iy == 0 || iy == 15) {
				int pos;
				if(ix == 0 || ix == 15)
					pos = iy % 4;
				else
					pos = ix % 4;
				switch(pos) {
				case 0: return 0.84f;
				case 1: return 1.28f;
				case 2: return 1.12f;
				case 3: return 0.72f;
				}
				return 0.0f; // can't happen
			} else if(ix == 1 || ix == 14 || iy == 1 || iy == 14)
				return 0.7f;
			else if(ix == 2 || ix == 3 || ix == 13 || ix == 12 || iy == 2 || iy == 3 || iy == 13 || iy == 12)
				return 1.0f;
			else if(ix == 4 || ix == 11 || iy == 4 || iy == 11)
				return 0.7f;
			else
				return 0.85f;
			
		} else if(ef == Effect.BRICKS || ef == Effect.SMALL_BRICKS) {
			float scalefacx, scalefacy;
			
			if(ef == Effect.SMALL_BRICKS) {
				y = (y*2)%h;
				x = (x*2)%w;
			}
			
			if(y < h/2) {
				scalefacx = (x + 0.5f) / (float)w;
				scalefacy = (y + 0.5f) / (float)h;
			} else {
				scalefacx = ((x + w/2) % w + 0.5f) / (float)w;
				scalefacy = (y - h/2 + 0.5f) / (float)h;
			}
			
			float a = 1/16f;
			
			if(ef == Effect.SMALL_BRICKS)
				a *= 2;
			
			float scalefac;
			if(scalefacx > 1-a || scalefacy >= 0.5f-a)
				scalefac = nedgefac;
			else if(scalefacx < a || scalefacy < a)
				scalefac = edgefac;
			else
				scalefac = 1.0f;
			return scalefac;

		} else if(ef == Effect.CARVED_BRICKS) {
			float x_ = (x+0.5f)/w, y_=(y+0.5f)/h;
			/*if(x_ >= 6/16f && x_ <= 10/16f && y_ >= 6/16f && y_ <= 10/16f) {
				if(x_ >= 9/16f || y_ >= 9/16f)
					return edgefac;
				return nedgefac;
			}
			if(x_ >= 4/16f && x_ <= 12/16f && y_ >= 4/16f && y_ <= 12/16f)
				return 1;*/
			int xi = (int)(x_*16);
			int yi = (int)(y_*16);
			float xid = Math.abs(x_ - 0.5f), yid = Math.abs(y_ - 0.5f);
			int i = Math.min((xid > yid ? xi : yi), (xid > yid ? 15-xi : 15-yi));
			boolean lowerright = (x_+y_)>=1;
			boolean equal = xi+yi == 15;
			switch(i)
			{
			case 0: return lowerright ? nedgefac : edgefac;
			case 1: return 1;
			case 2: return lowerright ? edgefac : nedgefac;
			case 3: return equal ? 1 : lowerright ? nedgefac : edgefac;
			case 4: return 1;
			case 5: return 1;
			case 6: return lowerright ? edgefac : nedgefac;
			case 7: return nedgefac;
			default:
				return i/16.0f;
			}
			
		} else
			throw new AssertionError(""+ef);
	}

	private static void blur(BufferedImage i, int xradius, int yradius) {
		
		int xr = xradius, yr = yradius;
		
		float[][][] cols = new float[i.getWidth()][i.getHeight()][];
		for(int x = 0; x < i.getWidth(); x++)
			for(int y = 0; y < i.getHeight(); y++)
				cols[x][y] = getcol(i, x, y);
		
		float[][][] newcols = new float[i.getWidth()][i.getHeight()][];
		
		for(int x = 0; x < i.getWidth(); x++) {
			for(int y = 0; y < i.getHeight(); y++) {
				float[] sum = {0, 0, 0};
				for(int dx = -xr; dx <= xr; dx++) {
					for(int dy = -yr; dy <= yr; dy++) {
						sum = add(sum, cols[(x+dx+i.getWidth())%i.getWidth()][(y+dy+i.getHeight())%i.getHeight()]);
					}
				}
				newcols[x][y] = mul(sum, 1.0f / ((xr*2+1)*(yr*2+1)));
				setcol(i, x, y, newcols[x][y]);
			}
		}
		
		
	}

	private static float calcNaturalVariance(BufferedImage i, float[] avgcol) {
		float max = 0;
		for(int x = 0; x < i.getWidth(); x++) {
			for(int y = 0; y < i.getHeight(); y++) {
				float[] icol = getcol(i, x, y);
				max = Math.max(max, Math.max(Math.max(icol[0]/avgcol[0], icol[1]/avgcol[1]), icol[2]/avgcol[2]));
			}
		}
		return max;
	}
	
	public static BufferedImage doMossyEffect1(BufferedImage i) {
		
		BufferedImage mossSeed = load("mossy");
		
		int w = i.getWidth(), h = i.getHeight();
		
		boolean[][] moss = new boolean[w][h];
		float[][] gradient = new float[w][h];
		
		for(int x = 0; x < w; x++)
			for(int y = 0; y < h; y++) {
				if((mossSeed.getRGB(x*mossSeed.getWidth()/w, y*mossSeed.getHeight()/h) & 0xFF000000) != 0)
					moss[x][y] = true;
				
				float[] px = getcol(i,x,y);
				gradient[x][y] = px[0]+px[1]+px[2]; 
			}
		
		for(int rep = 0; rep < 50; rep++) {
			for(int x = 0; x < w; x++)
				for(int y = 0; y < h; y++)
					if(moss[x][y]) {
						// gradient descent
						if(gradient[x][y] > gradient[(x+1)%w][y] && !moss[(x+1)%w][y]) {
							moss[(x+1)%w][y] = true;
							moss[x][y] = false;
							continue;
						}
						if(gradient[x][y] > gradient[(x+w-1)%w][y] && !moss[(x+w-1)%w][y]) {
							moss[(x+w-1)%w][y] = true;
							moss[x][y] = false;
							continue;
						}
						if(gradient[x][y] > gradient[x][(y+1)%h] && !moss[x][(y+1)%h]) {
							moss[x][(y+1)%h] = true;
							moss[x][y] = false;
							continue;
						}
						if(gradient[x][y] > gradient[x][(y+h-1)%h] && !moss[x][(y+h-1)%h]) {
							moss[x][(y+h-1)%h] = true;
							moss[x][y] = false;
							continue;
						}
					}
		}
		
		for(int x = 0; x < w; x++)
			for(int y = 0; y < h; y++)
				if(moss[x][y]) {
					float[] px = getcol(i,x,y);
					float v = (px[0]+px[1]+px[2])/3;
					px[0] = v/2;
					px[1] = v;
					px[2] = v/3;
					setcol(i,x,y, px);
				}
		
		return i;
	}

	public static BufferedImage doEffect(BufferedImage i, Effect effect) {
		switch(effect) {
		case BRICKS:
		case CARVED_BRICKS:
		case PILLAR:
		case PILLAR_HORIZONTAL:
			return doBricksEffect(i, 0.55f, 2.5f, 0, effect);
		case PILLAR_END:
			return doBricksEffect(i, 0.55f, 2.5f, 1, effect);
		case LARGE_TILES:
		case SMALL_BRICKS:
		case SMALL_TILES:
		case ORNATE_LARGE_TILES:
			return doBricksEffect(i, 0.65f, 1.6f, 0, effect);
		}
		throw new RuntimeException("unknown effect "+effect);
	}
}
